# $Id: Core.pm,v 1.11 2011/12/17 10:34:41 koba Exp $
#	Copyright(c) 2007-2011 Satoshi Kobayashi, All rights reserved.

package XiuMai::Module::Core;

use strict;
use XiuMai::Module;
use XiuMai::Text;
use XiuMai::Resource;
use XiuMai::Util qw(cdata);
use vars qw($VERSION @ISA @INLINE_MODULE @BLOCK_MODULE @NP_MODULE);

$VERSION = do { my @r = (q$Revision: 1.11 $ =~ /\d+/g);
				sprintf("%d." . "%02d" x $#r, @r) };

@ISA = qw(XiuMai::Module);

@INLINE_MODULE = qw( img color size br clear class tag );
@BLOCK_MODULE  = qw( import title contents include img footnote clear
					 class style icon lang tag );
@NP_MODULE     = qw( style );

sub import {
	my $self = shift;
	my (undef, $param) = @_;
	return $self->text->import_module(split(/,\s*/, $param));
}

sub title {
	my $self = shift;
	my (undef, $param) = @_;

	my $value = $self->text->block;

	my $title = $param;
	my $curr_title = $self->text->title;
	$title =~ s/\$\$/$curr_title/;
	$self->text->{title} = $title;

	return $value;
}

sub contents {
	my $self = shift;
	my $toc_num = @{$self->text->{toc}};

	my $value = $self->text->block;

	my @toc = @{$self->text->{toc}};
	splice(@toc, 0, $toc_num);

	return qq(<div class="contents">\n)
			. XiuMai::Text->new->parse(join("\n", @toc))
			. qq(</div>\n)
			. $value;
}

sub include {
	my $self = shift;
	return if (! $self->request);

	my (undef, $param) = @_;
	my $res = new XiuMai::Resource($self->request, $param)
			or return qq(<p class="error">)
						.cdata("$param: No such file or directory")
						.qq(</p>\n);
	$res = $res->_open
			or return qq(<p class="error">)
						.cdata("$param: Permission denied")
						.qq(</p>\n);
	my @src = split(/\x0d\x0a|\x0d|\x0a/, $res->text);
	unshift(@{$self->text->{SRC}}, @src);
	return '';
}

sub img {
	my $self = shift;
	my ($mode, $param) = @_;
	my ($src, $alt, @option) = split(/,\s*/, $param);

	return if ($src =~ /^javascript:/i);
	$src = cdata($self->_url($src));
	$alt = cdata($alt);

	my (%style, $nolink);
	foreach (@option) {
		if    ($_ eq 'nolink')			{ $nolink = 1; 				}
		elsif (/^(left|right)$/)		{ $style{float}   = $1;		}
		elsif (/^(top|middle|bottom)$/) { $style{'vertical-align'}
														  = $1;		}
		elsif (/^b(\d+)$/)				{ $style{border}  = "$1px";	}
		elsif (/^w(\d+)$/)				{ $style{width}   = "$1px";	}
		elsif (/^h(\d+)$/)				{ $style{height}  = "$1px";	}
		elsif (/^(\d+)x(\d+)$/)			{ $style{width}   = "$1px";
										  $style{height}  = "$2px";	}
		elsif (/^(.+):\s*(.*)$/)		{ $style{$1}      =  $2;	}
	}
	my $style;
	foreach (keys %style) {
		$style .= ' '	if (defined $style);
		$style .= cdata("$_: $style{$_};");
	}
	$style = defined $style ? qq(style="$style") : '';

	my $html = qq(<img src="$src" alt="$alt" $style/>);
	$html = qq(<a href="$src">$html</a>)	if (! $nolink);
	if ($mode eq 'block') {
		$html = qq(<div>$html</div>\n);
	}
	return $html;
}

sub color {
	my $self = shift;
	my (undef, $param, $value) = @_;
	return if (! defined $param || ! defined $value);
	my ($color, $bgcolor) = split(/,\s*/, $param);
	return if (! defined $color);
	my $style = cdata("color: $color;");
	$style .= cdata(" background-color: $bgcolor;")	if (defined $bgcolor);
	return qq(<span style="$style">$value</span>);
}

sub size {
	my $self = shift;
	my (undef, $param, $value) = @_;
	return if (! defined $param || ! defined $value);
	return if ($param !~ /^(:?\d+(:?px|%)|\d+(:?\.\d)?em)$/);
	my $size = cdata($param);
	return qq(<span style="font-size: $size;">$value</span>);
}

sub br {
	my $self = shift;
	return qq(<br />);
}

sub footnote {
	my $self = shift;
	return $self->text->footnote;
}

sub clear {
	my $self = shift;
	my ($mode, $param) = @_;
	my $clear = (defined $param && $param =~ /^(left|right)$/) ? $1 : 'both';
	return $mode eq 'inline'
				? qq(<br style="clear: $clear;" />)
				: qq(<div style="clear: $clear;"></div>\n);
}

sub class {
	my $self = shift;
	my ($mode, $param, $value) = @_;
	return if (! defined $param || ! defined $value);

	$param = cdata($param);
	return $mode eq 'inline'
				? qq(<span class="$param">$value</span>)
				: qq(<div class="$param">$value</div>\n);
}

sub style {
	my $self = shift;
	my (undef, $param, $value) = @_;

	if (defined $param) {
		my ($href, $media) = split(/,\s*/, $param, 2);
		$self->text->stylesheet(
				{ href => $self->_url($href), media => $media } );
	}
	elsif (defined $value) {
		$value =~ s/&/&amp;/g;
		$value =~ s/</&lt;/g;
		$value =~ /(.*)/s	and $value = $1;
		$self->text->style(qq(<style type="text/css">\n$value</style>\n));
	}
	else {
		return;
	}
	return '';
}

sub icon {
	my $self = shift;
	my (undef, $param) = @_;
	return if (! defined $param);
	$self->text->icon($self->_url($param));
	return '';
}

sub lang {
	my $self = shift;
	my (undef, $param) = @_;
	return if (! defined $param);
	$self->text->lang($param);
	return '';
}

sub tag {
	my $self = shift;
	my ($mode, $param, $value) = @_;
	return if (! defined $param);
	
	my ($tag, $attr) = $param =~ /^(\w+)((?:\s+\w+="[^"]*")*)$/	or return;
	$tag =~ /^SCRIPT$/i		and return;
	$tag =~ /^NOSCRIPT$/i	and return $value;
	$tag =~ /^(?:IFRAME|OBJECT|APPLET|EMBED)$/i		and return;

	my $new_attr = '';
	foreach ($attr =~ /\s+(\w+="[^"]*")/g) {
		my ($key, $val) = /^(\w+)="(.*)"$/;
		$val =~ s/&lt;/</g;
		$val =~ s/&gt;/>/g;
		$val =~ s/&quot;/"/g;
		$val =~ s/&amp;/&/g;

		return if ($key =~ /^ON/i);
		return if ($key =~ /^(?:SRC|HREF|ACTION)$/i && $val =~ /^javascript:/i);

		$new_attr .= qq( $key=").cdata($val).qq(");
	}

	return ($mode eq 'inline')
				? (defined $value)
					? qq(<$tag$new_attr>$value</$tag>)
					: qq(<$tag$new_attr />)
				: (defined $value)
					? qq(<$tag$new_attr>\n$value</$tag>\n)
					: qq(<$tag$new_attr />\n);
}

1;
