# $Id: Text.pm,v 1.40 2011/12/17 11:43:11 koba Exp $
#	Copyright(c) 2006-2011 Satoshi Kobayashi, All rights reserved.

package XiuMai::Text;

use strict;
use XiuMai::Util qw(cdata canonpath);
use vars qw($VERSION);

$VERSION = do { my @r = (q$Revision: 1.40 $ =~ /\d+/g);
				sprintf("%d." . "%02d" x $#r, @r) };

my $link_regex = qr{
		(?:https?|ftp)://[\w\.\-\/\~\?=%&#:;,\$\+]+
	|	(?:mailto:)?[\w\.\-\?\/]+@[\w\-\.]+
}x;

my $markup_regex;
if (5.006 <= $]) {
	$markup_regex
		= qr/ (?: (?: \'\'\' (?: (??{$markup_regex}) | . )*? \'\'\' )
				| (?:   \'\' (?: (??{$markup_regex}) | . )*? \'\'   )
				| (?:   \%\% (?: (??{$markup_regex}) | . )*? \%\%   )
				| (?:   \(\( (?: (??{$markup_regex}) | . )*? \)\)   )
				| (?:   \[\[ (?!\[)                    .  *? \]\]   )
				| (?:   \{\{                           .  *? \}\}   )
				| (?: &\w+
					  (?: \( .*? (?<!\\)\) )?
					  (?: { (?: (??{$markup_regex}) | . )*? } )? ;  ) )
			/x;
}
else {
	my $m1 		  = qr/(?:(?: '''       .*?''' )
						| (?:  ''       .*?''  )
						| (?:  %%       .*?%%  )
						| (?:\(\(       .*?\)\))
						| (?:  {{       .*?}}  )
						| (?:\[\[ (?!\[).*?\]\])
						|&\w+(?:\(.*?(?<!\\)\))?(?:{.*?})?;)/x;
	my $m2		  = qr/(?:(?: '''(?:$m1|.)*?''' )
						| (?:  ''(?:$m1|.)*?''  )
						| (?:  %%(?:$m1|.)*?%%  )
						| (?:\(\((?:$m1|.)*?\)\))
						| (?:  {{       . *?}}  )
						| (?:\[\[ (?!\[). *?\]\])
						|&\w+(?:\(.*?(?<!\\)\))?(?:{(?:$m1|.*)*?})?;)/x;
	$markup_regex = qr/(?:(?: '''(?:$m2|.)*?''' )
						| (?:  ''(?:$m2|.)*?''  )
						| (?:  %%(?:$m2|.)*?%%  )
						| (?:\(\((?:$m2|.)*?\)\))
						| (?:  {{       . *?}}  )
						| (?:\[\[ (?!\[). *?\]\])
						|&\w+(?:\(.*?(?<!\\)\))?(?:{(?:$m2|.*)*?})?;)/x;
}

sub new {
	my $class = shift;
	my ($res) = @_;
	return bless {
				Resource	=> $res,

				eof			=> [],
				toc			=> [],
				footnote	=> [],
				note_base	=> 0,

				lang		=> undef,
				stylesheet	=> [],
				icon		=> undef,
				style		=> '',

				module		=> {},
			}, $class;
}

sub resource {
	my $self = shift;
	return $self->{Resource};
}

sub request {
	my $self = shift;
	return if (! $self->{Resource});
	return $self->{Resource}->request;
}

sub style {
	my $self = shift;
	return $self->{style}			if (! @_);
	$self->{style} .= shift;
}

sub stylesheet {
	my $self = shift;
	return @{$self->{stylesheet}}	if (! @_);
	push(@{$self->{stylesheet}}, shift);
}

sub icon {
	my $self = shift;
	return $self->{icon}			if (! @_);
	$self->{icon} = shift;
}

sub lang {
	my $self = shift;
	return $self->{lang}			if (! @_);
	$self->{lang} = shift;
}

sub parse {
	my $self = shift;
	my ($src) = @_ ? shift : $self->{Resource}->text;
	my @src = split(/\x0d\x0a|\x0d|\x0a/, $src);
	$self->{SRC} = \@src;
	my $html = $self->import_module('Core', '_default');
	$html .= $self->block;
	$html .= $self->footnote;
	$self->{module} = {};
	return $html;
}

sub import_module {
	my $self = shift;
	my ($class, $prefix) = @_;
	$class =~ /^\w+$/	or return;
	$prefix ||= $class;
	eval qq(
		require XiuMai::Module::$class;
		\$self->{module}->{$prefix} = new XiuMai::Module::$class(\$self);
	);
	if ($@) {
		return qq(<p class="error">)
				.cdata(qq(Module "$class" not found.))
				.qq(</p>\n);
	}
	return '';
}

sub eof {
	my $self = shift;
	return unshift(@{$self->{eof}}, shift)		if (@_);
	return @{$self->{SRC}} == 0
			|| @{$self->{eof}} && $self->{SRC}->[0] eq $self->{eof}->[0];
}

sub reset {
	my $self = shift;
	shift @{$self->{SRC}}	if (@{$self->{SRC}});
	shift @{$self->{eof}}	if (@{$self->{eof}});
}

sub read_line {
	my $self = shift;
	return ! $self->eof && shift @{$self->{SRC}};
}

sub next_line {
	my $self = shift;
	return ! $self->eof && $self->{SRC}->[0];
}

sub block {
	my $self = shift;
	my ($eof) = @_;
	$self->eof($eof)	if (defined $eof);
	my $html = '';
	while (! $self->eof) {
		my $line = $self->read_line;
		next if ($line =~ /^$/);

		if    ($line =~ /^\!/)		{ $html .= $self->title	($line); 		}
		elsif ($line =~ /^\*/)		{ $html .= $self->h		($line, 1);		}
		elsif ($line =~ /^\-+$/)	{ $html .= $self->hr	($line); 		}
		elsif ($line =~ /^\s/)		{ $html .= $self->pre	($line);		}
		elsif ($line =~ /^\-/)		{ $html .= $self->li	($line, 'ul');	}
		elsif ($line =~ /^\+/)		{ $html .= $self->li	($line, 'ol');	}
		elsif ($line =~ /^\:/)		{ $html .= $self->dl	($line);		}
		elsif ($line =~ /^\|/)		{ $html .= $self->table	($line);		}
		elsif ($line =~ /^\#/)		{ $html .= $self->block_module($line);	}
		elsif ($line eq '>>')		{ $html .= $self->blockquote;			}
		elsif ($line eq '>|')		{ $html .= $self->block_pre;			}
		elsif ($line =~ /^\/\//)	{ 										}
		elsif ($line eq '/*')		{ 		   $self->np_block('*/');		}
		else						{ $html .= $self->p		($line); 		}
	}
	$self->reset		if (defined $eof);
	return $html;
}

sub np_block {
	my $self = shift;
	my ($eof) = @_;
	$self->eof($eof)	if (defined $eof);
	my $html = '';
	while (! $self->eof) {
		$html .= $self->read_line.qq(\n);
	}
	$self->reset		if (defined $eof);
	return $html;
}

sub title {
	my $self = shift;
	return $self->{title}	if (! @_);
	my ($line) = @_;
	$line =~ s/^!\s*//;
	$self->{title} = $self->simple_line($line,1);
	return qq(\n<h1>).$self->line($line,1).qq(</h1>\n);
}

sub p {
	my $self = shift;
	my ($line) = @_;

	my $align = '';
	if    ($line =~ s/^<//)	{ $align = qq( style="text-align: left;");		}
	elsif ($line =~ s/^=//)	{ $align = qq( style="text-align: center;");	}
	elsif ($line =~ s/^>//)	{ $align = qq( style="text-align: right;");		}
	else					{ $line =~ s/^~//;								}
	my $html = qq(<p$align>).$self->line($line);
	while (! $self->eof) {
		$line = $self->next_line;
		last if ($line =~ /^$/);
		last if ($line =~ /^\/\*$/);
		last if ($line =~ /^\/\//);
		last if ($line =~ /^[\~\<\=\>\!\*\s\-\+\:\|\#]/);
		$line = $self->read_line;

		$html .= $self->line($line);
	}
	chomp $html;
	return $html.qq(</p>\n);
}

sub h {
	my $self = shift;
	my ($line, $base_level) = @_;

	$line =~ s/^(\*{1,5})\s*//;
	my $level = length($1);

	my $id;
	$line =~ s/\s*\[([^\[\]]*)\]$//	and $id = $1;

	my $toc = $self->{toc};
	defined $id or $id = sprintf("SEC.%d", scalar @$toc + 1);

	my $text = $self->simple_line($line,1);

	push(@$toc, ('-' x $level).qq([[$text>#$id]]));

	$level += $base_level;
	$id = cdata($id);
	return qq(\n<h$level id="$id">).$self->line($line,1).qq(</h$level>\n);
}

sub hr {
	my $self = shift;
	my ($line) = @_;
	return qq(<hr />\n);
}

sub pre {
	my $self = shift;
	my ($line) = @_;

	$line =~ s/^ //;
	my $html = qq(<pre>).cdata($line).qq(\n);
	while (! $self->eof) {
		$line = $self->next_line;
		last if ($line !~ /^\s/);
		$line = $self->read_line;

		$line =~ s/^ //;
		$html .= cdata($line).qq(\n);
	}
	chomp $html;
	return $html.qq(</pre>\n);
}

sub li {
	my $self = shift;
	my ($line, $uo, $level) = @_;

	$level++;
	my $html = qq(<$uo>\n);
	$line =~ /^(\-+)\s*(.*)$/	if ($uo eq 'ul');
	$line =~ /^(\++)\s*(.*)$/	if ($uo eq 'ol');
	$html .= ($level < length($1))
				? qq(<li style="display: inline; list-style-type: none;">)
					.$self->li($line, $uo, $level)
				: qq(<li>).$self->line($2);
	while (! $self->eof) {
		$line = $self->next_line;
		last if ($line =~ /^\-*$/);
		last if ($line =~ /^\/\*$/);
		last if ($line =~ /^\/\//);
		last if ($line =~ /^[\~\<\=\>\!\*\s\:\|\#]/);

		if ($line =~ /^(\-+)\s*(.*)$/) {
			last if ($uo eq 'ul' && length($1) < $level);
			last if ($uo eq 'ol' && length($1) <= $level);
			$line = $self->read_line;
			if ($level < length($1)) {
				$html .= $self->li($line, 'ul', $level);
			}
			else {
				chomp $html;
				$html .= qq(</li>\n<li>).$self->line($2);
			}
		}
		elsif ($line =~ /^(\++)\s*(.*)$/) {
			last if ($uo eq 'ol' && length($1) < $level);
			last if ($uo eq 'ul' && length($1) <= $level);
			$line = $self->read_line;
			if ($level < length($1)) {
				$html .= $self->li($line, 'ol', $level);
			}
			else {
				chomp $html;
				$html .= qq(</li>\n<li>).$self->line($2);
			}
		}
		else {
			$line = $self->read_line;
			$html .= $self->line($line);
		}
	}
	chomp $html;
	return $html.qq(</li>\n</$uo>\n);
}

sub dl {
	my $self = shift;
	my ($line) = @_;

	my $html = qq(<dl>\n);
	$line =~ s/^://;
	$html .= qq(<dt>).$self->line($line,1).qq(</dt>\n);
	while (! $self->eof) {
		$line = $self->next_line;
		last if ($line =~ /^$/);
		last if ($line =~ /^\/\*$/);
		last if ($line =~ /^\/\//);
		last if ($line =~ /^[\~\<\=\>\!\*\s\-\+\|\#]/);
		$line = $self->read_line;

		if ($line =~ s/^://) {
			if (length($line)) {
				$html .= qq(<dt>).$self->line($line,1).qq(</dt>\n);
			}
		}
		else {
			$html .= $self->dd($line);
		}
	}
	return $html.qq(</dl>\n);
}

sub dd {
	my $self = shift;
	my ($line) = @_;

	my $html = qq(<dd>).$self->line($line);
	while (! $self->eof) {
		$line = $self->next_line;
		last if ($line =~ /^$/);
		last if ($line =~ /^\/\*$/);
		last if ($line =~ /^\/\//);
		last if ($line =~ /^[\~\<\=\>\!\*\s\-\+\:\|\#]/);
		$line = $self->read_line;

		$html .= $self->line($line);
	}
	chomp $html;
	return $html.qq(</dd>\n);
}

sub table {
	my $self = shift;
	my ($line) = @_;

	my $table = [];

	$line =~ s/^\|//;
	$line =~ s/\|$//;
	push(@{$table}, [ split(/\|/, $line, -1) ]);
	while (! $self->eof) {
		$line = $self->next_line;
		last if ($line !~ /^\|/);
		$line = $self->read_line;

		$line =~ s/^\|//;
		$line =~ s/\|$//;
		push(@{$table}, [ split(/\|/, $line, -1) ]);
	}

	my $html = qq(<table>\n);
	while (@{$table}) {
		my $tr = shift @{$table};
		my $col = 0;

		$html .= qq(<tr>);
		while (@{$tr}) {
			$col++;
			my $cell = shift @{$tr};
			next if ($cell =~ /^[\-\^]/);

			my $span = '';
			my $colspan = 1;
			foreach (@{$tr}) {
				last if (! /^-/);
				$colspan++;
			}
			$span .= qq( colspan="$colspan")	if ($colspan > 1);

			my $rowspan = 1;
			foreach (@{$table}) {
				last if (defined $_->[$col -1] && $_->[$col -1] !~ /^\^/);
				$rowspan++;
			}
			$span .= qq( rowspan="$rowspan")	if ($rowspan > 1);

			my $tag = ($cell =~ s/^~//) ? 'th' : 'td';

			my $align = '';
			if    ($cell =~ s/^=//) {
				$align = qq( style="text-align: center;");
			}
			elsif ($cell =~ s/^>//) {
				$align = qq( style="text-align: right;");
			}
			elsif ($cell =~ s/^<//) {
				$align = qq( style="text-align: left;");
			}

			$html .= qq(<$tag$span$align>).$self->line($cell,1).qq(</$tag>);
		}
		$html .= qq(</tr>\n);
	}
	return $html.qq(</table>\n);
}

sub blockquote {
	my $self = shift;
	my $html = qq(<blockquote>\n);
	$html .= $self->block('<<');
	return $html.qq(</blockquote>\n);
}

sub block_pre {
	my $self = shift;
	my $html = qq(<pre>);
	$html .= cdata($self->np_block('|<'));
	chomp $html;
	return $html.qq(</pre>\n);
}

sub simple_line {
	my $self = shift;
	my $line = $self->line(@_);

	$line =~ s/<.*?>//g;
	$line =~ s/&lt;/</g;
	$line =~ s/&gt;/>/g;
	$line =~ s/&quot;/"/g;
	$line =~ s/&amp;/&/g;

	return $line;
}

sub line {
	my $self = shift;
	my ($line, $no_eol) = @_;

	my $eol = $no_eol ? '' : ($line =~ s/(~)$// ? qq(<br />\n) : qq(\n));
	return $self->inline($line).$eol;
}

sub inline {
	my $self = shift;
	my ($inline) = @_;

	my @elem = $inline =~ /((?:(?!$markup_regex).)*)($markup_regex)?/g;
	my $html;
	while (@elem) {
		my $text = shift @elem;
		$html .= $self->text($text)		if (defined $text);
		my $markup = shift @elem;
		$html .= $self->markup($markup)	if (defined $markup);
	}
	return $html;
}

sub markup {
	my $self = shift;
	my ($markup) = @_;

	$markup =~ /^'''(.*)'''$/		and return qq(<strong>)
												.$self->inline($1)
												.qq(</strong>);
	$markup =~ /^''(.*)''$/			and return qq(<em>)
												.$self->inline($1)
												.qq(</em>);
	$markup =~ /^%%(.*)%%$/			and return qq(<del>)
												.$self->inline($1)
												.qq(</del>);
	$markup =~ /^\(\((.*)\)\)$/		and return $self->footnote($1);
	$markup =~ /^\[\[(.*)\]\]$/		and return $self->bracket($1);
	$markup =~ /^\{\{(.*)\}\}$/		and return cdata($1);

	return $self->module($markup);
}

sub bracket {
	my $self = shift;
	$_ = shift;
	my ($url, $text) = /^(.*)>(.*)$/ ? ($2, $1) : ($_, $_);
	if ($url =~ m|^/|) {
		$url = ($self->request ? $self->request->xiumai_base_url : '')
			 . canonpath($url);
	}
	$url = ''	if ($url =~ /^javascript:/i);
	$url = cdata($url);
	return qq(<a href="$url">).$self->inline($text).qq(</a>);
}

sub footnote {
	my $self = shift;
	if (@_) {
		push(@{$self->{footnote}}, shift);
		my $n  = @{$self->{footnote}};
		my $nb = $self->{note_base} || '';
		return qq(<sup class="footnote">)
				.qq(<a id="NOTEREF$nb.$n" href="#FOOTNOTE$nb.$n">*$n</a></sup>);
	}
	if (@{$self->{footnote}}) {
		my $n;
		my $nb = $self->{note_base} || '';
		my $html = qq(<div class="footnote"><hr /><ol>\n);
		foreach my $footnote (@{$self->{footnote}}) {
			$n++;
			$html .= qq(<li><sup><a id="FOOTNOTE$nb.$n" href="#NOTEREF$nb.$n">)
				   . qq(^</a></sup> ).$self->inline($footnote).qq(</li>\n);
		}
		$self->{footnote} = [];
		$self->{note_base}++;
		return $html.qq(</ol></div>\n);
	}
	return '';
}

sub module {
	my $self = shift;
	my ($module) = @_;

	my ($name, $param, $value)
			= $module =~ /^&(\w+)(?:\((.*?)\))?(?:{(.*)})?;$/;
	$param =~ s/\\(.)/$1/g		if (defined $param);

	my $html = $self->{module}->{_default}
				->dispatch('inline', $name, $param, $value);
	if (! defined $html) {
		$html = qq(<span class="error">);
		$html .= cdata("&$name");
		$html .= cdata("($param)")	if (defined $param);
		$html .= qq({).$self->inline($value).qq(})
									if (defined $value);
		$html .= qq(;</span>);
	}

	return $html;
}

sub block_module {
	my $self = shift;
	my ($module) = @_;

	my ($name, $param, $eof)
			= $module =~ /^#(\w+)(?:\((.*?)(?<!\\)\))?(?:<<(.*))?$/;
	return $self->p($module)		if (! defined $name);
	$param =~ s/\\(.)/$1/g		if (defined $param);

	my $html = $self->{module}->{_default}
					->dispatch('block', $name, $param, $eof);

	if (! defined $html) {
		$html = $self->p($module);	chomp $html;
		return qq(<div class="error">$html</div>\n);
	}

	return $html
}

sub text {
	my $self = shift;
	my ($text) = @_;

	my @elem = $text =~ /((?:(?!$link_regex).)*)($link_regex)?/g;
	my $html;
	while (@elem) {
		$html .= cdata(shift @elem);
		my $link = shift @elem;
		next	if (! defined $link);
		my $value = cdata($link);
		my $param = ($link =~ /:/) ? $value : cdata("mailto:$link");
		$html .= qq(<a href="$param">$value</a>);
	}
	return $html;
}

1;
